###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.quantlib.black_scholes import opt_BS
from onyx.quantlib.tree_pricers import opt_BT
from onyx.quantlib.montecarlo_pricers import MC_vanilla, MC_american
from onyx.econometrics.gbm_functions import GbmPricePaths

import numpy as np
import math
import unittest

# --- conversions to daily for MC paths
to_daily = 1.0 / 255.0


###############################################################################
class UnitTest(unittest.TestCase):

    def test_vanilla_convergence(self):
        # --- seed the random generator to make unittest deterministic
        np.random.seed(145)

        # --- price paths parameters
        spot, strike, vol, rd, rf, n_days = (75.0, 100.0,
                                             0.25, 0.05, 0.03, 60)

        payoff = lambda s: max(s - strike, 0.0)

        term = n_days*to_daily
        rd_daily = rd*to_daily
        rf_daily = rf*to_daily

        bs_prem = opt_BS("C", spot, strike,
                         vol, term, rd, rf, ret_type="P")

        for n_sims in (1000, 5000, 10000, 50000, 100000, 500000, 1000000):
            paths = spot*GbmPricePaths(vol*math.sqrt(to_daily), n_days,
                                       rd_daily, rf_daily, n_sims, True)

            mc_prem, mc_err = MC_vanilla(paths, payoff, rd, rf, term)

#            print(n_sims, mc_prem, mc_err, bs_prem)

    # -------------------------------------------------------------------------
    def test_american_convergence(self):
        # --- seed the random generator to make unittest deterministic
        np.random.seed(145)

        # --- price paths parameters
        spot, strike, vol, rd, rf, n_days = (75.0, 100.0,
                                             0.25, 0.05, 0.03, 60)

        payoff = lambda s: max(s - strike, 0.0)

        term = n_days*to_daily
        rd_daily = rd*to_daily
        rf_daily = rf*to_daily

        bt_prem = opt_BT("C", spot, strike, vol, term,
                         rd, rf, "AMERICAN", n_steps=4080)

        for n_sims in (1000, 5000, 10000, 50000, 100000, 250000):
            paths = spot*GbmPricePaths(vol*math.sqrt(to_daily), n_days,
                                       rd_daily, rf_daily, n_sims, False)

            mc_prem, mc_err = MC_american(paths, payoff, term, rd, rf,
                                          basis="poly", opt_type="C",
                                          spot=spot, strike=strike, vol=vol)

#            print(n_sims, mc_prem, mc_err, bt_prem)

#    # -------------------------------------------------------------------------
#    def test_vanilla(self):
#        # --- seed the random generator to make unittest deterministic
#        np.random.seed(123)
#
#        # --- price paths parameters
#        spot, vol, rd, n_sims = (100.0, 0.25, 0.05, 10000)
#
#        # --- test two scenarios for rf (i.e. smaller and larger than rd)
#        for rf in (0.03, 0.07):
#            rd_daily = rd*to_daily
#            rf_daily = rf*to_daily
#
#            for n_days in (60, 120, 255):
#                term = n_days*to_daily
#                paths = spot*GbmPricePaths(vol*math.sqrt(to_daily), n_days,
#                                           rd_daily, rf_daily, n_sims, True)
#
#                for strike in (75.0, 100, 125):
#                    # --- payoff function for a call option
#                    payoff = lambda s: max(s - strike, 0.0)
#
#                    mc_prem, mc_err = MC_vanilla(paths, payoff, rd, rf, term)
#
#                    bs_prem = opt_BS("C", spot, strike,
#                                     vol, term, rd, rf, ret_type="P")
#
#                    delta = abs(mc_prem - bs_prem)
#
#                    # --- chack that estimated MC premium is compatible with
#                    #     the BS one within 2 sigma (not very stringent).
#                    self.assertLessEqual(delta / mc_err, 2.0)

    # -------------------------------------------------------------------------
#    def test_american(self):
#        # --- seed the random generator to make unittest deterministic
#        np.random.seed(145)
#
#        # --- price paths parameters
#        spot, vol, rd, n_sims = (100.0, 0.25, 0.05, 5000)
#
#        # --- conversions to daily for MC paths
#        to_daily = 1.0 / 255.0
#
#        # --- test two scenarios for rf (i.e. smaller and larger than rd)
#        for rf in (0.03, 0.07):
#            rd_daily = rd*to_daily
#            rf_daily = rf*to_daily
#
#            for n_days in (60, 120, 255):
#                term = n_days*to_daily
#                paths = spot*GbmPricePaths(vol*math.sqrt(to_daily), n_days,
#                                           rd_daily, rf_daily, n_sims, False)
#
#                for strike in (75.0, 100, 125):
#                    # --- payoff function for a call option
#                    payoff = lambda s: max(s - strike, 0.0)
#
#                    mc_prem, mc_err = MC_american(paths, payoff, term, rd, rf,
#                                                  basis="poly", opt_type="C",
#                                                  spot=spot, strike=strike, vol=vol)
#
#                    bt_prem = opt_BT("C", spot, strike, vol, term,
#                                     rd, rf, "AMERICAN", n_steps=4080)
#
#                    delta = abs(mc_prem - bt_prem)
#
#                    # --- chack that estimated MC premium is compatible with
#                    #     the BS one within 2 sigma (not very stringent).
#                    self.assertLessEqual(delta / mc_err, 2.0)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, CalcTerm
from onyx.core import SelectField, ReferenceField, DictField, Archivable
from onyx.core import GetVal, GraphNodeVt, MktIndirectionFactory, EvalBlock

from ..quantlib.black_scholes import implied_vol

import bisect

__all__ = ["EquityVolMatrix"]


###############################################################################
class EquityVolMatrix(Archivable):
    """
    EquityVolMatrix class.
    """
    Asset = ReferenceField(obj_type="EquityAsset")
    Style = SelectField(options=["EUROPEAN", "AMERICAN"])

    # -------------------------------------------------------------------------
    def __post_init__(self):
        exch_code = GetVal(GetVal(self.Asset, "Exchange"), "Code")
        self.Name = ("EQ VOLMAT {0:s} "
                     "{1:s}").format(self.Asset.Symbol, exch_code)

    # -------------------------------------------------------------------------
    @MktIndirectionFactory(DictField)
    def PrcMarks(self, graph):
        """
        Price Marks are structured as follows:
        {
            expiry: [
                {
                    strike: [price_call, price_put],
                    ...
                }
                ...
            ]
        }
        NB: these are NOT used for pricing, but only to calculate implied
            volatilities that are then stored in VolMarks.
        """
        pass

    # -------------------------------------------------------------------------
    @MktIndirectionFactory(DictField)
    def VolMarks(self, graph):
        """
        Marks are structured as follows:
        {
            expiry: [
                {
                    strike: [vol_call, vol_put],
                    ...
                }
                ...
            ]
        }
        """
        pass

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def ExpirationDates(self, graph):
        return sorted([Date.parse(d) for d in graph(self, "VolMarks")])

    # -------------------------------------------------------------------------
    @GraphNodeVt()
    def StrikesByExpDate(self, graph, exp_date):
        exp_date = exp_date.isoformat()
        return sorted([float(k) for k in graph(self, "VolMarks")[exp_date]])

    # -------------------------------------------------------------------------
    @GraphNodeVt()
    def VolByExpStrike(self, graph, exp_date, strike):
        exp_date_str = exp_date.isoformat()
        strike_str = str(strike)
        vols = graph(self, "VolMarks")[exp_date_str][strike_str]

        return 0.5*(vols[0] + vols[1])

    # -------------------------------------------------------------------------
    @GraphNodeVt()
    def VolByExpStrikeType(self, graph, exp_date, strike, opt_type):
        # --- use only the first letter capitalized
        opt_type = opt_type[0].upper()

        if strike == "ATM":
            strike = graph(self, "AtmStrikeByExpDate", exp_date)

        exp_date_str = exp_date.isoformat()
        strike_str = str(strike)
        vols = graph(self, "VolMarks")[exp_date_str][strike_str]

        if opt_type == "C":
            return vols[0]
        elif opt_type == "P":
            return vols[1]
        else:
            raise NameError("Unrecognized option type, {0:s}".format(opt_type))

    # -------------------------------------------------------------------------
    @GraphNodeVt()
    def AtmVol(self, graph, exp_date, opt_type):
        return graph(self, "VolByExpStrikeType", exp_date, "ATM", opt_type)

    # -------------------------------------------------------------------------
    @GraphNodeVt()
    def AtmStrikeByExpDate(self, graph, exp_date):
        spot = graph(graph(self, "Asset"), "Spot")
        strikes = graph(self, "StrikesByExpDate", exp_date)

        ih = bisect.bisect_left(strikes, spot)
        il = max(0, ih - 1)
        idx = il if spot - strikes[il] < strikes[ih] - spot else ih

        return strikes[idx]

    # -------------------------------------------------------------------------
    def generate_vol_marks(self, pd, price_marks):

        vol_marks = price_marks
        is_american = (self.Style == "AMERICAN")

        with EvalBlock() as eb:
            eb.change_value("Database", "MktDataDate", pd)
            spot = GetVal(self.Asset, "Spot")

        for exp_date_str, opts_by_strike in price_marks.items():
            exp_date = Date.parse(exp_date_str)

            rd = ...
            rf = GetVal(self.Asset, "DividendYield", pd, exp_date)

            term = CalcTerm(pd, exp_date)

            for strike_str, (p_call, p_put) in opts_by_strike.items():
                strike = float(strike_str)

                iv_c = implied_vol("C", p_call, spot, strike,
                                   term, rd, rf, american=is_american)
                iv_p = implied_vol("P", p_put, spot, strike,
                                   term, rd, rf, american=is_american)

                vol_marks[exp_date_str][strike_str] = [iv_c, iv_p]

        return vol_marks



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, RDate, Curve, GCurve
from onyx.core import Interpolate, ApplyAdjustments
from onyx.core import Archivable, ValueType, GetVal, EvalBlock
from onyx.core import StringField, ListField, ReferenceField
from onyx.core import MktIndirectionFactory, EnforceArchivableEntitlements

import numpy as np
import collections

__all__ = ["CorporateActions"]


###############################################################################
@EnforceArchivableEntitlements("Database", "ArchivedOverwritable")
class CorporateActions(Archivable):
    """
    Class used to represent corporate actions, archived by their Declared Date.
    """
    Asset = ReferenceField(obj_type="EquityAsset")

    # -------------------------------------------------------------------------
    @MktIndirectionFactory(StringField)
    def DvdCcy(self, graph):
        """
        Here we store the currency used for dividend payments.
        """
        pass

    # -------------------------------------------------------------------------
    @MktIndirectionFactory(ListField)
    def ActionInfo(self, graph):
        """
        Here we store the list of corporate actions as of the relevant ex-date.
        The structure is the following:
        [
            {
                "ex-date": first date after which shareholders are no longer
                           entitled to the benefits of the coporate action,
                "declared-date": the date when the corporate action is
                                 officially annaunced,
                "pay-date": the date when the payment takes place,
                "type": ["cash", "scrip", "split", "rights issue", "other"],
                "amount": for cash it's the gross dividend amount,
                          for scrip, split, rights issue, and "other" it's the
                          multiplicative factor.
            },
        ]
        """
        pass

    # -------------------------------------------------------------------------
    @ValueType()
    def LastKnot(self, graph):
#        date = date or graph("Database", "MktDataDate")
        date = graph("Database", "MktDataDate")
        return self.get_dated("ActionInfo", date, strict=False)

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def NextExDateActions(self, graph, start):
        # FIXME: this is extremely inefficient, we should use a specialized
        #        query
        pos_dt = graph("Database", "PositionsDate")
        crv = self.corp_actions_curve(start=max(start, pos_dt))
        if len(crv):
            return crv.front.value
        else:
            return [{"ex-date": Date.high_date(), "type": None}]

    # -------------------------------------------------------------------------
    @ValueType()
    def AdjustCurve(self, graph):
        ccy = graph(graph(self, "Asset"), "Denominated")
        mul = graph(graph(self, "Asset"), "Multiplier")

        def adjuster(crv):
            start = crv.front.date
            end = crv.back.date
            offset = RDate("-1b")
            adj_crv = Curve()

            for date, actions in self.corp_actions_curve(start=start, end=end):

                if date == start:
                    # --- remove this knot from the curve and skip this
                    #     corporate action
                    start += RDate("+1b")
                    continue

                prev_biz_date = date + offset
                prev_close = Interpolate(crv, prev_biz_date)[3]

                cash_items, mul_items = [], []
                for action in actions:
                    dvd_type = action["type"]
                    amount = action["amount"]
                    if dvd_type == "cash":
                        fx_adj = self.fx_ajustment(ccy, mul, date)
                        cash_items.append(amount*fx_adj)
                    else:
                        mul_items.append(amount)

                adj = 1.0 - sum(cash_items) / prev_close
                adj *= np.prod(mul_items)
                adj_crv[date] = adj

            adj_crv = Curve(adj_crv.dates,
                            np.cumprod(adj_crv.values[::-1])[::-1])

            return ApplyAdjustments(crv, adj_crv).crop(start=start, end=end)

        return adjuster

    # -------------------------------------------------------------------------
    def corp_actions_curve(self, by_date="ex-date", start=None, end=None):
        info_by_date = collections.defaultdict(list)
        if by_date == "ex-date":
            # --- given that archived records are stored by ex-date, we can use
            #     a specialized faster query
            for _, info in self.get_history("ActionInfo", start, end):
                for item in info:
                    info_by_date[item[by_date]].append(item)
        else:
            for _, info in self.get_history("ActionInfo"):
                for item in info:
                    date = item[by_date]
                    if date < start or date > end:
                        continue
                    info_by_date[date].append(item)

        return GCurve(info_by_date.keys(), info_by_date.values())

    # -------------------------------------------------------------------------
    def dvd_curve(self, start=None, end=None):
        """
        N.B.: we only include cash dividends here.
        """
        info_crv = self.corp_actions_curve(start=start, end=end)
        dvd_crv = Curve()
        for _, value in info_crv:
            for action in value:
                if action["type"] == "cash":
                    date = action["ex-date"]
                    if date in dvd_crv:
                        dvd_crv[date] += action["amount"]
                    else:
                        dvd_crv[date] = action["amount"]

        return dvd_crv

    # -------------------------------------------------------------------------
    def fx_adjustment(self, ccy, mul, mdd):
        with EvalBlock() as eb:
            eb.change_value("Database", "MktDataDate", mdd)
            dvd_ccy = GetVal(self, "DvdCcy")

        if ccy == dvd_ccy:
            return 1.0
        else:
            # --- dvd_ccy -> ccy
            to_usd = GetVal("{0:3s}/USD".format(dvd_ccy), "SpotByDate", mdd)
            to_ccy = 1.0 / GetVal("{0:3s}/USD".format(ccy), "SpotByDate", mdd)

            return to_usd * to_ccy / mul


# -----------------------------------------------------------------------------
def prepare_for_test():
    from ..system import ufo_database
    ufo_database.prepare_for_test()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from apscheduler.executors.pool import ProcessPoolExecutor
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore

from . import scheduling_fns

import tornado.httpserver
import tornado.ioloop
import tornado.web

import argh
import logging
import os

# --- configure logging
LOG_FMT = "%(asctime)-15s %(levelname)-8s %(name)-38s %(message)s"
logging.basicConfig(level=logging.INFO, format=LOG_FMT)


# -----------------------------------------------------------------------------
def run(port=9100, sqlite_path=None):

    sqlite_base = os.getenv("HOME", "./")
    sqlite_path = sqlite_path or os.path.join(sqlite_base, ".jobs.sqlite")
    sqlite_url = "sqlite:///{0:s}".format(sqlite_path)

    jobs_stores = {
        "default": SQLAlchemyJobStore(url=sqlite_url),
    }

    jobs_executors = {
        "default": ProcessPoolExecutor(2)
    }

    handlers = [
        (r"/jobs$", scheduling_fns.JobsHandler),
        (r"/jobs/(\w+$)", scheduling_fns.JobsHandler),
    ]

    app = scheduling_fns.SchedulingServer(
        jobs_stores,
        jobs_executors,
        handlers,
    )

    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(port)
    tornado.ioloop.IOLoop.instance().start()


# -----------------------------------------------------------------------------
def main():
    argh.dispatch_command(run)


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    run()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .optimization import QuickMin, ConvergenceError, Hessian_FD_g

from math import log as m_log, sqrt as m_sqrt, pi as m_pi

import numpy as np
import scipy.optimize as scipyopt

__all__ = ["Garch11", "IGarch11", "Garch11_CondVol"]

EPS = 1.e-6
MAXITER = 500


################################################################################
class QuickMinGarch11(QuickMin):
    def enforce_constraints(self):
        # --- parameters are positive
        self.pos = np.where(self.pos > EPS, self.pos, EPS)
        # --- domain constratint
        if self.pos[1] + self.pos[2] >= 1.0:
            self.pos[1:] /= (self.pos[1] + self. pos[2] + 1.e-2)


# -----------------------------------------------------------------------------
def Garch11(ret, skip=10, guess=None, verbose=False):
    """
    Description:
        ...
    Inputs:
        x - ...
    Returns:
        ...
    """
    if guess is None:
        guess = np.array([1.e-4, 0.15, 0.80])

    esq = (ret - ret.mean())**2
    args = (esq, skip)
    bounds = ( ( EPS, None ), ( EPS, 1.0 - EPS ), ( EPS, 1.0 - EPS ) )

    qm = QuickMinGarch11(guess, Garch11_NegLogLikelihood_g,
                         args=args, dt=0.1, gthr=1.e-8, verbose=verbose)

    for k, parms in enumerate(qm):
        if verbose:
            loglike = Garch11_NegLogLikelihood(parms, esq, skip)
            print(k, loglike, parms)
        if k > MAXITER:
            msg = """
            Garch11: QuickMin couldn't converge after {0;d} iterations. Last
            estimated parameters are: {1!s}""".format(k, parms)
            raise ConvergenceError(msg)

    loglike = Garch11_NegLogLikelihood(parms, esq, skip)

    covmat = np.linalg.inv(Hessian_FD_g(Garch11_NegLogLikelihood_g,
                                        parms, esq, skip))

    if np.isnan(covmat).any():
        raise ConvergenceError("Garch11, NaNs in the covariance matrix")

    return parms, -loglike, covmat


# -----------------------------------------------------------------------------
def GARCH11_grid_search(o_range, a_range, b_range, args, verbose=False):
    g_min = None 
    g_parms = None

    for o0 in o_range:
        z = np.zeros((b_range.shape[0], a_range.shape[0]))
        for j, a0 in enumerate(a_range):
            for i, b0 in enumerate(b_range):
                if a0 + b0 >= 1.0:
                    z[i,j] = 0.0
                else:
                    parms = np.array([o0, a0, b0])
                    z[i,j] = Garch11_NegLogLikelihood(parms, *args)
    
        zmin = z.min()
        bmin, amin = np.where(z == zmin)

        if verbose:
            print("minimum: {0:.2f} at {1:f},{2:f},"
                  "{3:f}".format(zmin, o0, a_range[amin], b_range[bmin]))

        g_min = min(g_min, zmin) or zmin
        g_parms = [o0, a_range[amin],
                       b_range[bmin]] if g_min == zmin else g_parms

    if verbose:
        print("global minimum is {0:.2f} at {1:f},{2:f},"
              "{3:f}".format(g_min, g_parms[0], g_parms[1], g_parms[2]))

    return g_min, np.array(g_parms)


# -----------------------------------------------------------------------------
def Garch11_CondVar(omega, alpha, beta, ret, skip):
    """
    Description:
        ...
    Inputs:
        x - ...
    Returns:
        ...
    """
    esq = (ret - ret.mean())**2
    n = len(esq)
    var = np.zeros_like(esq)
    var[0] = esq[skip:].std()

    for i in range(1, n):
        var[i] = omega + alpha*esq[i-1] + beta*var[i-1]

    return var[skip:]

# -----------------------------------------------------------------------------
def Garch11_CondVol(omega, alpha, beta, ret, skip):
    return np.sqrt(Garch11_CondVar(omega, alpha, beta, ret, skip))


# -----------------------------------------------------------------------------
def Garch11_Returns(omega, alpha, beta, vol0, ndays):
    var = vol0*vol0
    ret = np.zeros(ndays)

    for i, z in enumerate(np.random.standard_normal(ndays)):
        ret[i] = m_sqrt(var)*z
        var = omega + (alpha*z*z + beta)*var

    return ret


# -----------------------------------------------------------------------------
def Garch11_Fcast(omega, alpha, beta,
                  ret, skip, ndays, nsims=1000, covmat=None):

    if covmat is None:
        omega *= np.ones(nsims)
        alpha *= np.ones(nsims)
        beta *= np.ones(nsims)
    else:
        z = np.random.multivariate_normal((omega, alpha, beta), covmat, nsims)

        omega = z[:,0]
        alpha = z[:,1]
        beta = z[:,2]

        den = np.where(alpha + beta >= 0.999, alpha + beta + 0.001, 1.0)

        alpha /= den
        beta  /= den

    esq = (ret - ret.mean())**2
    var = esq[skip:].std()*np.zeros(nsims)
    esq = np.tile(esq, (nsims, 1))

    for i in range(1, esq.shape[1]):
        var = omega + alpha*esq[:,i-1] + beta*var

    ret = np.zeros((nsims, ndays))

    if nsims % 2 == 0:
        def randn(n):
            z = np.random.standard_normal(n / 2)
            return np.hstack((-z, z))
    else:
        def randn(n):
            return np.random.standard_normal(n)

    for i in range(ndays):
        z = randn(nsims)
        ret[:,i] = np.sqrt(var)*z
        var = omega + (alpha*z*z + beta)*var

    return np.squeeze(ret)


# ------------------------------------------------------------------------------
def Garch11_NegLogLikelihood(parms, esq, skip):
    # --- unpack parameters
    omega, alpha, beta = parms

    n = len(esq)
    var = esq[skip:].std()

    for i in range(1, skip):
        var = omega + alpha*esq[i-1] + beta*var

    loglike = -0.5*(n - skip)*m_log(2.0*m_pi)

    for i in range(skip, n):
        var = omega + alpha*esq[i-1] + beta*var
        loglike -= 0.5*(m_log(var) + esq[i] / var)

    return -loglike


# ------------------------------------------------------------------------------
def Garch11_NegLogLikelihood_g(parms, esq, skip):
    # --- unpack parameters
    omega, alpha, beta = parms

    n = len(esq)
    var = esq[skip:].std()

    for i in range(1, skip):
        var = omega + alpha*esq[i-1] + beta*var

    grad = np.zeros(3)

    g0 = 0.0
    g1 = 0.0
    g2 = 0.0

    for i in range(skip, n):
        g0 = 1.0 + beta*g0
        g1 = esq[i-1] + beta*g1
        g2 = var + beta*g2

        var = omega + alpha*esq[i-1] + beta*var
        coeff = (1.0 - esq[i] / var) / var

        grad += coeff*np.array([g0, g1, g2])

    return 0.5*grad


# ------------------------------------------------------------------------------
def IGarch11(ret, skip=10):
    """
    Description:
        ...
    Inputs:
        x - ...
    Returns:
        ...
    """
    args = ((ret - ret.mean())**2, 10)
    bounds = [0.0, 1.0]
    res = scipyopt.minimize_scalar(IGarch11_NegLogLikelihood,
                                   args=args, method="bounded", bounds=bounds)

    if not res["success"]:
         raise ConvergenceError("IGarch11: linear "
                                "optimization couldn't converge")

    return np.array([0.0, 1.0 -res["x"], res["x"]]), -res["fun"]


# ------------------------------------------------------------------------------
def IGarch11_NegLogLikelihood(beta, esq, skip):
    n = len(esq)
    var = esq[skip:].std()

    for i in range(1, skip):
        var = (1.0 - beta)*esq[i-1] + beta*var

    loglike = -0.5*(n - skip)*m_log(2.0*m_pi)

    for i in range(skip, n):
        var = (1.0 - beta)*esq[i-1] + beta*var
        loglike -= 0.5*(m_log(var) + esq[i] / var)

    return -loglike

################################################################################
#  RegTest
# ------------------------------------------------------------------------------
if __name__ == "__main__":

#    parms = [ 1.e-5, 0.10, 0.85 ]
#    ret = Garch11_Returns( *parms, vol0 = 0.01, ndays = 100, nsims = 1000 )

    parms = [ 1.e-4, 0.10, 0.85, 0.02 ]
    ret = Garch11_Returns( *parms, ndays = 1000 )

    parms, _, covmat = Garch11( ret, verbose = False )

    z = np.random.multivariate_normal( ( 0.0, 0.0, 0.0 ), covmat, 10000 )

    err_distr = lambda p, z: 1.0 / ( 1.0 + ( 1.0 - p ) / p * np.exp( z ) )

    o = err_distr( parms[0], z[:,0] )
    a = err_distr( parms[1], z[:,1] )
    b = err_distr( parms[2], z[:,2] )




###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GetVal, OnyxTestCase
from onyx.system import ufo_holiday_calendar

import unittest


###############################################################################
class UnitTest(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        ufo_holiday_calendar.prepare_for_test()

    # -------------------------------------------------------------------------
    def test_Holidays(self):
        self.assertEqual(GetVal("USD_CAL", "Holidays"), set())


if __name__ == "__main__":
    from onyx.core import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, AddObj, GetObj, UpdateObj
from onyx.core import ObjDbQuery, ObjDbTransaction, ObjNotFound, ValueType
from onyx.core import CreateInMemory, UfoBase, GetVal, SetVal, PurgeObj
from onyx.core import DateField, ReferenceField, FloatField, StringField
from onyx.core import SelectField, unique_id

from .tradable_api import AddByInference

import collections
import random
import logging

__all__ = [
    "Transaction",
    "TransactionError",
    "TransactionWarning",
    "CommitTransaction",
    "RollbackTransaction",
    "DeleteTransaction",
    "AgeTransaction",
    "TransactionsBy"
]

logger = logging.getLogger(__name__)

# --- templates for common queries

QUERY_INSERT_POSEFF = """
INSERT INTO PosEffects VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s);"""

QUERY_DELETE_POSEFF = "DELETE FROM PosEffects WHERE Transaction=%s;"


# -----------------------------------------------------------------------------
def create_poseffects_table():
    with ObjDbTransaction("PosEffects Table Create"):
        ObjDbQuery("""
CREATE TABLE PosEffects (
  TradeDate   date NOT NULL,
  TimeCreated timestamp with time zone NOT NULL,
  Book        character varying(64) NOT NULL,
  Qty         double precision NOT NULL,
  Unit        character varying(64) NOT NULL,
  UnitType    character varying(64) NOT NULL,
  Transaction character varying(64) NOT NULL,
  Event       character varying(64) NOT NULL,
  Trade       character varying(64) NOT NULL );

CREATE INDEX pos_by_trade_date ON PosEffects (TradeDate);
CREATE INDEX pos_by_time_created ON PosEffects (TimeCreated);
CREATE INDEX pos_by_book ON PosEffects (Book);
CREATE INDEX pos_by_trade ON PosEffects (Trade);

CREATE OR REPLACE FUNCTION notify_trigger() RETURNS trigger AS $$
DECLARE
BEGIN
    IF (TG_OP = 'DELETE') THEN
        PERFORM pg_notify(TG_TABLE_NAME, OLD.Book);
        RETURN OLD;
    ELSIF (TG_OP = 'UPDATE') THEN
        PERFORM pg_notify(TG_TABLE_NAME, OLD.Book);
        PERFORM pg_notify(TG_TABLE_NAME, NEW.Book);
        RETURN NEW;
    ELSIF (TG_OP = 'INSERT') THEN
        PERFORM pg_notify(TG_TABLE_NAME, NEW.Book);
        RETURN NEW;
    END IF;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER PosEffects_Notify
    AFTER INSERT OR UPDATE OR DELETE ON PosEffects
    FOR EACH ROW EXECUTE PROCEDURE notify_trigger();""")


###############################################################################
class TransactionError(Exception):
    pass


###############################################################################
class TransactionWarning(Warning):
    pass


###############################################################################
Position = collections.namedtuple("Position",
                                  "TradeDate Book Qty "
                                  "Unit UnitType Name Event Trade")


###############################################################################
class Transaction(UfoBase):
    """
    Transactions represent trading events:
    - Transactions are the only means to change positions. Positions effects
      associated to each transaction are two sided and must always sume to
      zero.
    - Each transaction refers to one and only one trade.
    - Each transaction has marker to define the its type and a pointer to the
      previous transaction in the chain.
    - Each transaction has marker to indentify the event that gave rise to its
      creation.
    """
    TransactionDate = DateField()
    SecurityTraded = ReferenceField(obj_type="TradableObj")
    Quantity = FloatField(default=0.0)
    Party = ReferenceField(obj_type="Book")
    Counterparty = ReferenceField(obj_type="Book")
    Trade = ReferenceField(obj_type="Trade")
    PrevTransaction = ReferenceField(obj_type="Transaction")
    Event = StringField()
    Marker = SelectField(options=["HEAD", "AMENDED", "AGED",
                                  "BACKOUT", "ROLLEDBACK", "DELETED"])

    # -------------------------------------------------------------------------
    def __post_init__(self):
        self.Name = self.Name or Transaction.random_name()

    # -------------------------------------------------------------------------
    @ValueType()
    def Denominated(self, graph):
        return graph(graph(self, "SecurityTraded"), "Denominated")

    # -------------------------------------------------------------------------
    @ValueType()
    def Trader(self, graph):
        return graph(graph(self, "Trade"), "Trader")

    # -------------------------------------------------------------------------
    @ValueType()
    def TransactionInfo(self, graph):
        """
        Return a dictionary {attribute: value}, including only the stored
        attributes that can be altered by a transaction amendment.
        """
        attrs = self._json_fields.copy()
        attrs = sorted(attrs.difference({"Trade", "Event", "Marker"}))
        return dict(zip(attrs, [graph(self, attr) for attr in attrs]))

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        qty = graph(self, "Quantity")
        sec = graph(self, "SecurityTraded")
        return qty*graph(sec, "Leaves")

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        mtm = 0.0
        for leaf, qty in graph(self, "Leaves").items():
            mtm += qty*graph(leaf, "MktValUSD")
        return mtm

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        ccy = graph(self, "Denominated")
        spot_fx = graph("{0:3s}/USD".format(ccy), "Spot")
        return graph(self, "MktValUSD") / spot_fx

    # -------------------------------------------------------------------------
    @ValueType()
    def PositionEffects(self, graph):
        sec = graph(self, "SecurityTraded")
        qty = graph(self, "Quantity")
        date = graph(self, "TransactionDate")
        name = graph(self, "Name")
        event = graph(self, "Event")
        trade = graph(self, "Trade")
        cpty1 = graph(self, "Party")
        cpty2 = graph(self, "Counterparty")

        pos_effects = [
            Position(date, cpty1, qty,
                     sec, graph(sec, "ObjType"), name, event, trade),
            Position(date, cpty2, -qty,
                     sec, graph(sec, "ObjType"), name, event, trade)
        ]

        return pos_effects

    # -------------------------------------------------------------------------
    def delete(self):
        if GetVal("Database", "TradesDeletable"):
            query = """DELETE FROM PosEffects WHERE Transaction=%s;
                       NOTIFY PosEffects, %s; NOTIFY PosEffects, %s;"""
            ObjDbQuery(query, parms=(self.Name, self.Party, self.Counterparty))
        else:
            raise TransactionError("Trying to delete a "
                                   "transaction without permission")

    # -------------------------------------------------------------------------
    @classmethod
    def random_name(cls):
        return "TmpTrs-{0:>07d}".format(random.randrange(0, 10000000, 1))

    # -------------------------------------------------------------------------
    @classmethod
    def format_name(cls, date, trd_id):
        return "TRS {0:s} {1:s}".format(date.strftime("%Y%m%d"), trd_id)


# -------------------------------------------------------------------------
def CommitTransaction(transaction, stored_name=None):
    """
    Description:
        Insert/amend a transaction, taking care of all position effects.
    Inputs:
        transaction - the instance of Transaction to commit
        stored_name - the name of the corresponding stored transaction, used
                      for amendments
    Returns:
        The Transaction instance.
    """
    stored_name = stored_name or transaction.Name

    # --- clone security traded: AddByInference is needed to generate the
    #     proper ImpliedName.
    sec = AddByInference(GetObj(transaction.SecurityTraded).clone())

    # --- get transaction info before reloading the transaction
    info = GetVal(transaction, "TransactionInfo")
    info["SecurityTraded"] = sec.Name

    # --- need to overwrite transaction name AFTER looking up TransactionInfo,
    #     otherwise we end up  using TransactionInfo of the stored transaction
    transaction.Name = stored_name

    try:
        # --- reload security: this is needed to avoid basing decisions on
        #     attribute values that might have been set in memory
        transaction = GetObj(transaction.Name, refresh=True)

    except ObjNotFound:
        # --- transaction not found, it must be added to the backend along with
        #     its position effects
        transaction.SecurityTraded = sec.Name
        transaction.Marker = "HEAD"
        return insert_transaction(transaction)

    else:
        return amend_transaction(transaction, info)


# -----------------------------------------------------------------------------
def insert_transaction(transaction):
    """
    This function should never be called directly. Use TransactionCommit
    instead.
    """
    if transaction.Marker != "HEAD":
        raise TransactionError("Only head transaction can be "
                               "created. Marker for {0:s} is "
                               "{1:s}.".format(transaction.Name,
                                               transaction.Marker))

    # --- create a new transaction to make sure that Version, ChangedBy, and
    #     timestamps are fresh
    name = transaction.format_name(Date.today(), unique_id(8))
    transaction = transaction.clone(Name=name)
    time_created = transaction.TimeCreated

    logger.info("creating transaction {0:s}".format(name))

    with ObjDbTransaction("Insert", "SERIALIZABLE"):
        AddObj(transaction)

        for pos in GetVal(transaction, "PositionEffects"):
            ObjDbQuery(QUERY_INSERT_POSEFF,
                       parms=(pos.TradeDate, time_created, pos.Book,
                              pos.Qty, pos.Unit, pos.UnitType, pos.Name,
                              pos.Event, pos.Trade))

    return transaction


# -----------------------------------------------------------------------------
def amend_transaction(head, info):
    """
    This function should never be called directly. Use TransactionCommit
    instead.
    """
    if head.Marker != "HEAD":
        raise TransactionError("Only head transaction can be "
                               "amended. Marker for {0:s} is "
                               "{1:s}.".format(head.Name, head.Marker))

    # --- need to amend?
    same = True
    for attr, value in info.items():
        same = (value == getattr(head, attr))
        if not same:
            break
    if same:
        raise TransactionWarning("found nothing to amend "
                                 "for transaction {0:s}".format(head.Name))

    if head.TimeCreated >= Date.today():
        # --- same day amendment: execute in-place update
        for attr, value in info.items():
            SetVal(head, attr, value)

        logger.info("same day amedment "
                    "for transaction {0:s}".format(head.Name))

        time_created = head.TimeCreated
        with ObjDbTransaction("Amend", "SERIALIZABLE"):
            UpdateObj(head)

            # --- delete existing position effects for head transaction
            ObjDbQuery(QUERY_DELETE_POSEFF, (head.Name, ))

            # --- insert new position effects
            for pos in GetVal(head, "PositionEffects"):
                ObjDbQuery(QUERY_INSERT_POSEFF,
                           parms=(pos.TradeDate, time_created,
                                  pos.Book, pos.Qty, pos.Unit, pos.UnitType,
                                  pos.Name, pos.Event, pos.Trade))

        transaction = head

    else:
        # --- previous day amendment: book a backout and then a new transaction
        logger.info("previous day amedment "
                    "for transaction {0:s}".format(head.Name))

        with ObjDbTransaction("Amend", "SERIALIZABLE"):
            head.Marker = "AMENDED"
            UpdateObj(head)

            backout = head.clone()
            backout.Name = head.format_name(Date.today(), unique_id(8))
            backout.Marker = "BACKOUT"
            backout.PrevTransaction = head.Name
            AddObj(backout)

            new = head.clone()
            new.Name = head.format_name(Date.today(), unique_id(8))
            new.Marker = "HEAD"
            new.PrevTransaction = backout.Name

            CreateInMemory(new)
            for attr, value in info.items():
                SetVal(new, attr, value)

            AddObj(new)

            # --- update position effects table:
            # ---  1) first the back-out transaction: quantities are flipped to
            #         offset original transaction
            for pos in GetVal(backout, "PositionEffects"):
                ObjDbQuery(QUERY_INSERT_POSEFF,
                           parms=(pos.TradeDate, backout.TimeCreated,
                                  pos.Book, -pos.Qty, pos.Unit, pos.UnitType,
                                  pos.Name, pos.Event, pos.Trade))

            # ---  2) then the new transaction with the amendments
            for pos in GetVal(new, "PositionEffects"):
                ObjDbQuery(QUERY_INSERT_POSEFF,
                           parms=(pos.TradeDate, new.TimeCreated,
                                  pos.Book, pos.Qty, pos.Unit, pos.UnitType,
                                  pos.Name, pos.Event, pos.Trade))

        transaction = new

    return transaction


# -----------------------------------------------------------------------------
def RollbackTransaction(head):
    """
    Description:
        Rolls back a transaction and all its position effects.
    Inputs:
        head - the transaction to roll-back
    Returns:
        None.
    """
    # --- reload security: this is needed to avoid basing decisions on
    #     attribute values that might have been set in memory
    head = GetObj(head.Name, refresh=True)

    if head.Marker != "HEAD":
        raise TransactionError("Only head transaction can be "
                               "rolled back. Marker for {0:s} is "
                               "{1:s}.".format(head.Name,
                                               head.Marker))

    if head.TimeCreated >= Date.today():
        # --- same day roll-back: there are two sub-cases
        if head.PrevTransaction is None:
            # --- case 1: delete transaction (all postion effectes are taken
            #             care of automatically)
            with ObjDbTransaction("Rollback", "SERIALIZABLE"):
                PurgeObj(head.Name)

        else:
            # --- case 2: delete last two transactions and re-mark the amended
            #             one as "HEAD"
            backout = GetObj(head.PrevTransaction, refresh=True)
            amended = GetObj(backout.PrevTransaction, refresh=True)

            # --- check that markers conform to expectation
            if head.Marker != "HEAD":
                raise TransactionError("Wrong Marker for transaction "
                                       "{0:s}: {1:s} instead of "
                                       "HEAD".format(head.Name,
                                                     head.Marker))
            if backout.Marker != "BACKOUT":
                raise TransactionError("Wrong Marker for transaction "
                                       "{0:s}: {1:s} instead of "
                                       "BACKOUT".format(backout.Name,
                                                        backout.Marker))
            if amended.Marker != "AMENDED":
                raise TransactionError("Wrong Marker for transaction "
                                       "{0:s}: {1:s} instead of "
                                       "AMENDED".format(amended.Name,
                                                        amended.Marker))
            amended.Marker = "HEAD"

            with ObjDbTransaction("Rollback", "SERIALIZABLE"):
                PurgeObj(head.Name)
                PurgeObj(backout.Name)
                UpdateObj(amended)

    else:
        # --- previous day roll-back: book backout transactions for both the
        #     current head and backout transactions
        with ObjDbTransaction("Rollback", "SERIALIZABLE"):
            # --- we start with a backout of the head transaction
            head.Marker = "ROLLEDBACK"
            UpdateObj(head)

            head_backout = head.clone()
            head_backout.Name = head.format_name(Date.today(), unique_id(8))
            head_backout.Marker = "BACKOUT"
            head_backout.PrevTransaction = head.Name
            AddObj(head_backout)

            # --- now update position effects: quantities are flipped to
            #     offset original transactions
            for pos in GetVal(head_backout, "PositionEffects"):
                ObjDbQuery(QUERY_INSERT_POSEFF,
                           parms=(pos.TradeDate, head_backout.TimeCreated,
                                  pos.Book, -pos.Qty, pos.Unit, pos.UnitType,
                                  pos.Name, pos.Event, pos.Trade))

            # --- if head was the result of an amendment, look up the backout
            #     transaction and make it the new head.
            #     NB: a backout transaction has the same sign as the amended
            #         transaction (although when the positions effects are
            #         saved the sign is flipped), therefore no need to flip it
            #         here.
            if head.PrevTransaction is not None:
                backout = GetObj(head.PrevTransaction, refresh=True)
                reverse = backout.clone()
                reverse.Name = backout.format_name(Date.today(), unique_id(8))
                reverse.Marker = "HEAD"
                reverse.PrevTransaction = head_backout.Name
                AddObj(reverse)

                for pos in GetVal(reverse, "PositionEffects"):
                    ObjDbQuery(QUERY_INSERT_POSEFF,
                               parms=(pos.TradeDate, reverse.TimeCreated,
                                      pos.Book, pos.Qty, pos.Unit,
                                      pos.UnitType, pos.Name, pos.Event,
                                      pos.Trade))


# -----------------------------------------------------------------------------
def DeleteTransaction(transaction_name):
    """
    Description:
        Delete a transaction and all its position effects.
    Inputs:
        transaction_name - the name of the transaction to delete
    Returns:
        None.
    """
    # --- reload security: this is needed to avoid basing decisions on
    #     attribute values that might have been set in memory
    transaction = GetObj(transaction_name, refresh=True)

    if transaction.Marker != "HEAD":
        raise TransactionError("Only head transaction can be "
                               "deleted. Marker for {0:s} is "
                               "{1:s}.".format(transaction.Name,
                                               transaction.Marker))

    if transaction.TimeCreated >= Date.today():
        # --- same day delete: recursively delete transaction-chain
        with ObjDbTransaction("Delete", "SERIALIZABLE"):
            prev = transaction.PrevTransaction
            PurgeObj(transaction)
            while prev is not None:
                transaction = prev
                prev = transaction.PrevTransaction
                PurgeObj(transaction)

    else:
        # --- previous day delete: book backout transaction
        with ObjDbTransaction("Delete", "SERIALIZABLE"):
            transaction.Marker = "DELETED"
            UpdateObj(transaction)

            backout = transaction.clone()
            backout.Name = transaction.format_name(Date.today(), unique_id(8))
            backout.Marker = "BACKOUT"
            backout.PrevTransaction = transaction.Name
            AddObj(backout)

            # --- this is a dummy head-transaction that is created to preserve
            #     the rule stating that every change to a transaction should be
            #     represented by zero (for same day) or two (for previous day)
            #     new transactions
            #     NB: the quantity is manually set to zero!
            dummy = transaction.clone()
            dummy.Name = transaction.format_name(Date.today(), unique_id(8))
            dummy.Quantity = 0.0  # important!!!
            dummy.Marker = "HEAD"
            dummy.PrevTransaction = backout.Name

            AddObj(dummy)

            # --- now update position effects: quantities are flipped to
            #     offset original transactions
            for pos in GetVal(backout, "PositionEffects"):
                ObjDbQuery(QUERY_INSERT_POSEFF,
                           parms=(pos.TradeDate, backout.TimeCreated,
                                  pos.Book, -pos.Qty, pos.Unit, pos.UnitType,
                                  pos.Name, pos.Event, pos.Trade))

            # --- save positions effects for the dummy head-transaction: we
            #     need this (even if Qty is zero) to be able to lookup the
            #     head transaction for a given trade (required by the method
            #     TransactionsStored of a Trade instance)
            for pos in GetVal(dummy, "PositionEffects"):
                ObjDbQuery(QUERY_INSERT_POSEFF,
                           parms=(pos.TradeDate, dummy.TimeCreated,
                                  pos.Book, pos.Qty, pos.Unit, pos.UnitType,
                                  pos.Name, pos.Event, pos.Trade))


# -----------------------------------------------------------------------------
def AgeTransaction(head, date):
    """
    Description:
        Age a transaction and all its position effects.
    Inputs:
        head - the transaction to roll-back
        date - the date of the ageing event
    Returns:
        None.
    """
    # --- reload security: this is needed to avoid basing decisions on
    #     attribute values that might have been set in memory
    head = GetObj(head.Name, refresh=True)

    if head.Marker != "HEAD":
        raise TransactionError("Only head transaction can be aged. Marker "
                               "for {0:s} is {1:s}.".format(head.Name,
                                                            head.Marker))

    sec = GetVal(head, "SecurityTraded")
    event = GetVal(sec, "ExpectedTransaction")
    securities = GetVal(sec, "ExpectedSecurities", event)

    with ObjDbTransaction("Ageing Event", "SERIALIZABLE"):
        head.Marker = "AGED"
        UpdateObj(head)

        backout = head.clone()
        backout.Name = head.format_name(Date.today(), unique_id(8))
        backout.Marker = "BACKOUT"
        backout.TransactionDate = date
        backout.PrevTransaction = head.Name
        AddObj(backout)

        for pos in GetVal(backout, "PositionEffects"):
            ObjDbQuery(QUERY_INSERT_POSEFF,
                       parms=(pos.TradeDate, backout.TimeCreated,
                              pos.Book, -pos.Qty, pos.Unit, pos.UnitType,
                              pos.Name, pos.Event, pos.Trade))

        # --- insert new transactions
        prev_trans = backout.Name

        for sec_info in securities:
            # --- add security
            sec = AddByInference(sec_info["Security"])
            # --- append new transaction
            insert_transaction(Transaction(
                TransactionDate=date,
                SecurityTraded=sec.Name,
                Quantity=head.Quantity*sec_info["Quantity"],
                Party=GetVal(head, "Party"),
                Counterparty=GetVal(head, "Counterparty"),
                Trade=head.Trade,
                PrevTransaction=prev_trans,
                Event=event,
                Marker="HEAD",
            ))

            # --- only the first of the new transactions points to the previous
            #     trnsaction in the chain (which one exactly of the new
            #     transactionsdoesn't matter as they would have to be all
            #     rolled back together)
            prev_trans = None


# -----------------------------------------------------------------------------
def TransactionsBy(book=None, trade=None,
                   security=None, date_range=None, head_only=True):
    """
    Description:
        Return all transactions matching the given criteria.
    Inputs:
        book       - a given book
        trade      - a given trade
        security   - a given security
        date_range - a tuple of start and end date
        head_only  - if True, only returns the Head trade
    Returns:
        A list of trade names.
    """
    extquery = []
    criteria = []

    if book is not None:
        extquery.append("Book=%s")
        criteria.append(book)

    if trade is not None:
        extquery.append("Trade=%s")
        criteria.append(trade)

    if security is not None:
        extquery.append("Unit=%s")
        criteria.append(security)

    if date_range is not None:
        extquery.append("TradeDate BETWEEN %s AND %s")
        criteria.append(date_range[0])
        criteria.append(date_range[1])

    if len(criteria):
        query = ("SELECT DISTINCT(Transaction) "
                 "FROM PosEffects WHERE {0:s};").format(" AND ".join(extquery))
        rows = ObjDbQuery(query, parms=criteria, attr="fetchall")

    else:
        query = "SELECT DISTINCT(Transaction) FROM PosEffects;"
        rows = ObjDbQuery(query, attr="fetchall")

    if head_only:
        return [row.transaction for row in rows
                if GetVal(row.transaction, "Marker") == "HEAD"]
    else:
        return [row.transaction for row in rows]


# -----------------------------------------------------------------------------
def prepare_for_test():
    import psycopg2
    try:
        create_poseffects_table()
    except psycopg2.ProgrammingError as err:
        if psycopg2.errorcodes.lookup(err.pgcode) == "DUPLICATE_TABLE":
            pass
        else:
            raise



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import numpy as np
import math

__all__ = ["AR1", "AR2", "ARMA11", "ARMA21"]


# -----------------------------------------------------------------------------
def AR1(x, const=False):
    """
    Description:
        Fit a timeseries to a AR(1) process and return key statistics.
    Inputs:
        x     - the input vector
        const - if True, use cointegration tables with a constant
    Returns:
        A tuple containing:
            - AR(1) root
            - stdandard error
            - critical value (based on MacKinnon tables 2010)
            - residuals
    """
    lx = np.copy(x[:-1])
    dx = x[1:] - lx
    lx -= lx.mean()

    rho = np.dot(dx, lx) / np.dot(lx, lx)
    res = dx - rho*lx

    n = len(res) - 1.0
    ni = 1.0 / n

    # --- standard error of the slope
    stderr = math.sqrt(np.dot(res, res) / np.dot(lx, lx) * ni)

    # --- critical value (1%) from MacKinnon tables 2010
    if const:
        cv = -3.43035 - 6.5393*ni - 16.786*ni*ni - 79.433*ni*ni*ni
    else:
        cv = -2.56574 - 2.2358*ni - 3.627*ni*ni

    return rho + 1.0, stderr, rho / stderr / cv, res


# -----------------------------------------------------------------------------
def AR2(x):
    """
    Description:
        Fit a timeseries to a AR(2) process and return key statistics.
    Inputs:
        x - the input vector
    Returns:
        ...
    """
    r1, r2 = np.linalg.lstsq(np.vstack([x[1:-1], x[:-2]]).T, x[2:])[0]

    res = x[2:] - r1*x[1:-1] - r2*x[:-2]

    return r1, r2, res


# -----------------------------------------------------------------------------
def ARMA11(x):
    """
    Description:
        Fit a timeseries to a ARMA(1,1) process and return key statistics.
    Inputs:
        x - the input vector
    Returns:
        ...
    """
    r = np.dot(x[:-1], x[1:]) / np.dot(x[:-1], x[:-1])
    e = x[1:] - r*x[:-1]

    root, phi = np.linalg.lstsq(np.vstack([x[1:-1], e[:-1]]).T, x[2:])[0]

    res = x[2:] - root*x[1:-1] - phi*e[:-1]

    return root, phi, res


# -----------------------------------------------------------------------------
def ARMA21(x):
    """
    Description:
        Fit a timeseries to a ARMA(2,1) process and return key statistics.
    Inputs:
        x - the input vector
    Returns:
        ...
    """
    r1, r2 = np.linalg.lstsq(np.vstack([x[1:-1], x[:-2]]).T, x[2:])[0]

    e = x[2:] - r1*x[1:-1] - r2*x[:-2]
    A = np.vstack([x[2:-1], x[1:-2], e[:-1]]).T

    r1, r2, phi = np.linalg.lstsq(A, x[3:])[0]

    res = x[3:] - r1*x[2:-1] - r2*x[1:-2] - phi*e[:-1]

    return r1, r2, phi, res


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    import pylab

    def generate_AR1(root, x0, eps):
        x = x0
        yield x
        for e in eps:
            x = root*x + e
            yield x

    root = 0.99
    ndays = 150
    nsims = 20000

    err = np.random.randn(ndays - 1)
    ar1 = np.fromiter(generate_AR1(root, 0.0, err), np.double, ndays)

    rho, stderr, cv, res = AR1(ar1)

    # --- standardize residuals
    res -= res.mean()
    res /= res.std()

    root = rho + 1.0

    roots = []
    for k in range(nsims):
        idx = np.random.randint(0, ndays - 1, ndays - 1)
        ar1 = np.fromiter(generate_AR1(root, ar1[0],
                                       res[idx]), np.double, ndays)
        rho, _, _, _ = AR1(ar1)
        roots.append(rho + 1.0)

    print(root, np.mean(roots), np.median(roots), np.percentile(roots, 99))

    pylab.hist(roots, 200)
    pylab.show()
